#include <libguile.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <iconv.h>
#include <locale.h>


const int mo_magic = 0x950412de;
const int mo_swapped_magic = 0xde120495 ;
const int mo_revision = 0;


void append_to(SCM *lst,const char *orig,const char *tr)
{
  *lst = scm_cons(scm_cons(scm_mem2string(orig,strlen(orig)),
			   scm_mem2string(tr,strlen(tr))),
		  *lst);
}


SCM scm_load_mo(SCM file)
{
  SCM_ASSERT(SCM_STRINGP(file), file, SCM_ARG1, "scm_load_mo");

  const char *filename = SCM_STRING_CHARS(file);
  SCM result = SCM_EOL;

  /* loaded *.mo data */
  int fd = 0;
  size_t fsize = -1;
  struct stat st;
  const char *mem = 0;
  int i=0;
  iconv_t conv = (iconv_t)(-1);

  /* file's data */
  int32_t magic = 0;
  int32_t revision = 0;
  int32_t number = 0;
  int32_t original = 0;
  int32_t translated = 0;

  if ((fd = open(filename,O_RDONLY)) < 0) return SCM_BOOL_F;

  if (fstat(fd,&st) < 0) return SCM_BOOL_F;
  
  if ((mem = mmap(NULL, st.st_size, PROT_READ, MAP_SHARED, fd, 0)) == MAP_FAILED)
      return SCM_BOOL_F;

  magic = *(int32_t*)(mem);
  revision = *(int32_t*)(mem+4);
  number = *(int32_t*)(mem+8);
  original = *(int32_t*)(mem+12);
  translated=*(int32_t*)(mem+16);

  if ((magic != mo_magic) && (magic != mo_swapped_magic)) return SCM_BOOL_F;
  if (revision != mo_revision) return SCM_BOOL_F;
  
 
  /* 'cause all original strings are in lexicographic order first string is "".
     It allows us to determine charset
  */
  {
    const char *first_string = (mem + *(int32_t*)(mem + original + 4));
    if (*first_string == 0) /* empty string */
      {
	const char *translation_data = (mem + *(int32_t*)(mem+translated + 4));
	const char *charset_info = strstr(translation_data,"charset=");

	if (charset_info)
	  {
	    /*extract charset*/
	    char *stop;
	    char *charset=strdup(charset_info+strlen("charset="));
	    stop = index(charset,'\n');
	    if (stop )*stop = 0;

	    /*setup iconv */
	    if (!strstr(charset,"utf")) conv = iconv_open("utf-8",charset); /* need conversion */
	    free(charset);
	  }
      }
  }


 /* walk throw destriptors of original strings */
  for(i=1;i<number;++i)
    {
      const char *orig_descriptor = (mem + original + 2*4*i);
      const char *tr_descriptor = (mem + translated + 2*4*i);

      //int32_t orig_str_size = *(int32_t*)(orig_descriptor);
      int32_t orig_str_offset = *(int32_t*)(orig_descriptor + 4);
      const char *orig_str = mem + orig_str_offset;

      //int32_t tr_str_size = *(int32_t*)(tr_descriptor);
      int32_t tr_str_offset = *(int32_t*)(tr_descriptor + 4);
      const char *tr_str = mem + tr_str_offset;

      //find also appropriate translated string
      if (conv != (iconv_t)(-1))
	{

	  size_t tr_len = strlen(tr_str);
	  char *tr = strdup(tr_str);
	  char *tr2 = tr; /* iconv change pointer location, so we have to made a copy */
	  size_t buf_len = tr_len*8; /* new message may be bigger */
	  char *buf = calloc(1,sizeof(char)*buf_len);
	  char *buf2 = buf; /* iconv change pointer location, so we have to made a copy */

	  if ( iconv(conv,&tr2,&tr_len,&buf2,&buf_len) != (size_t)(-1))
	    append_to(&result,orig_str,buf);
	  else
	    append_to(&result,orig_str,tr_str);
	  
	  free(tr);
	  free(buf);
	}
      else
	append_to(&result,orig_str,tr_str);
    }

  if (conv != (iconv_t)(-1)) iconv_close(conv);

  munmap((void*)mem,fsize);
  close(fd);
  return result;
}


void scm_init_gettext(void)
{
        setlocale(LC_ALL,"");
        scm_c_define_gsubr("load-mo", 1, 0, 0,
                           (SCM (*)())scm_load_mo);
}
